/* Control software for PIC ADC Logger by AKIZUKI DENSHI */
/* Copyright (c) 1999-2001 Noriaki Mitsunaga */
/* $Id: picadc.c,v 1.3 2001/06/09 06:11:36 jm3wud Exp jm3wud $ */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/types.h>

#define CR 13
#define LF 10
#define RETRY 5

#ifdef __FreeBSD__
#define DEVICE "/dev/cuaa0"
#endif

#ifdef __NetBSD__
#define DEVICE "/dev/ttyb"

#define NEED_USLEEP
#define NEED_NODELAY
#endif

#ifdef __LINUX
#define DEVICE "/dev/ttyS0"

#define NEED_USLEEP
#define NEED_NODELAY
#endif

#undef DEBUG

static struct termios savecon;
static int asy;

void ioinit()
{
	struct termios ttybuf;

	tcgetattr(0, &ttybuf);
	tcgetattr(0, &savecon);

	ttybuf.c_iflag &= ~(IXON | INLCR | ICRNL | ISTRIP);;
/* 	ttybuf.c_iflag |= (IGNCR); */
	ttybuf.c_lflag &= ~(ICANON|ISIG|ECHO);
	ttybuf.c_cflag &= ~(PARENB | CSIZE);
	ttybuf.c_cflag |= (CS8 | CRTSCTS);

	ttybuf.c_cc[VTIME] = '\0';
	ttybuf.c_cc[VMIN] = '\01';

	tcsetattr(0, TCSANOW, &ttybuf);
	return;
}

void ioreinit()
{
	tcsetattr(0, TCSANOW, &savecon);
	return;
}

int
asy_init(int fd, speed_t speed)
{
	struct termios	termios;


	tcgetattr(fd, &termios);

	termios.c_iflag = (IGNBRK | IGNPAR);
	termios.c_oflag = 0;
	termios.c_lflag = 0;
	termios.c_cflag = (CS8 | CREAD | CSTOPB | CLOCAL);
	cfsetispeed(&termios, speed);
	cfsetospeed(&termios, speed);
	termios.c_cc[VTIME] = 1; /* 0.1 sec */
	termios.c_cc[VMIN] = 0;

	if (tcsetattr(fd, TCSANOW, &termios) < 0) {
                perror("ioctl could not set parameters for IORser");
                return -1;
        }

        return 0;
}

int readchars(int fd, char buf[], size_t bufsiz)
{
  int i, r;

  buf[0] = 0;
  for(i=0; i<bufsiz; i++) {
#ifdef NEED_USLEEP
    usleep(10000);
#endif
    r = read(fd, &buf[i], 1);
    if(r <= 0) {
      break;
    }
    if(buf[i] == 0)
      break;
  }
  buf[i] = 0;

  return(i);
}

int printmode()
{
  char c = '?';
  char buf[BUFSIZ+1];
  int len = 0;
  int i;

  for(i=0; i<RETRY; i++) {
    write(asy, &c, 1);
/*     printf("Sent ?\n"); */
    len = readchars(asy, buf, BUFSIZ);
    if(strncmp(buf, "PIC-LOG V", 9) == 0)
      break;
    if(strncmp(buf, "PIC-ADC V", 9) == 0)
      break;
    if(strncmp(buf, "PIC-REALV", 9) == 0)
      break;
  }
  if(i == RETRY) {
    fprintf(stderr, "Retried but failed.\n");
    return(-1);
  } else 
    printf("%s", buf);
  return(0);
}

void sendbytes(int fd, char buf[], int num)
{
  int i;
  char c;
  int sum;

  sum = 0;
  for(i=0; i<num; i++) {
    usleep(10000); /* 0.1ms */
    write(fd, &buf[i], 1);
    sum = (sum + buf[i]) & 0xff;
  }

  c = (0x100 - sum) & 0xff;
  usleep(10000);
  write(fd, &c, 1);
}

int readchar(int fd)
{
  int i;
  unsigned char r;

  for(i=0; i<RETRY; i++) {
    if(read(fd, &r, 1) == 1)
      break;
#ifdef NEED_USLEEP
    usleep(10000);
#endif
  }
  if(i == RETRY) {
/*     fprintf(stderr, "Retry failed.\n"); */
    return(-1);
  }
/*   fprintf(stderr, "%d %d %c\n", i, r, r); */
  return(r);
}

int setbaudrate(int fd, int speed)
{
  char b9600[]  = {0x4f, 0xaa, 0x55, 0xaa, 0xaa, 0xaa};
  char b19200[] = {0x27, 0x55, 0x2a, 0x55, 0x55, 0x55};
  char s = 0;
  int i;

  if(speed == 9600) {
    s = 's';
    write(fd, &s, 1);
    if(readchar(fd) != s)
      return(-1);
    sendbytes(fd, b9600, 6);

    if(readchar(fd) != 'S')
      return(-2);

    asy_init(fd, B9600);
  }
  if(speed == 19200) {
    s = 's';
    write(fd, &s, 1);
    if(readchar(fd) != s)
       return(-1);
    sendbytes(fd, b19200, 6);

    if(readchar(fd) != 'S')
      return(-2);

    asy_init(fd, B19200);
  }
  return(0);
}

int setPCMODE(int fd)
{
  char s;
  int i;

  s = 't';
  for(i=0; i<RETRY; i++) {
    write(fd, &s, 1);
    if(readchar(fd) == s)
      break;
  }
  if(i==RETRY)
    return(-1);
  return(0);
}

int readDIN(int fd)
{
  char s;
  int c, i;

  s = 'i';
  write(fd, &s, 1);
  if(readchar(fd) != s) {
    fprintf(stderr, "Reading port B failed.\n");
    return;
  }

  c = readchar(fd);
  if(c<0)
    fprintf(stderr, "Reading port B failed.\n");
  else
    c = c >> 4;
  return(c);
}

int writeD(int fd, int data)
{
  char s;

  s = 'o';
  write(fd, &s, 1);
  if(readchar(fd) != s) {
    fprintf(stderr, "Writing port B failed.\n");
    return(-1);
  }
  write(fd, &data, 1);

  return(0);
}

void printDIN(int fd)
{
  int c;

  c = readDIN(fd);
  if(c>=0)
    printf("%d\n", c);
}

int readAIN(int fd, int ch, int ref)
{
  char s;
  int c;

  s = 'g';
  write(fd, &s, 1);
  if(readchar(fd) != s) {
    fprintf(stderr, "Reading AD failed(1).\n");
    return;
  }
  s = (0xc4 + ((ch & 0x3) << 3)) & 0xff;
  if(ref = 1)
    s += 1; /* XXX $B$&$^$/$$$C$F$$$J$$(B */
  write(fd, &s, 1);
  
  c = readchar(fd);
  if(c<0)
    fprintf(stderr, "Reading AD failed(2).\n");

  return(c);
}

void printAIN(int fd, int ch, int ref)
{
  printf("%d\n", readAIN(fd, ch, ref));
}

int readAINn(int fd, int ch, int ref, int num, unsigned char data[])
{
  int c, i;
  char s, buf[3];

  s = 'a';
  write(fd, &s, 1);
  if(readchar(fd) != s) {
    fprintf(stderr, "Reading AD failed(1).\n");
    return;
  }
  c = (0xc4 + ((ch & 0x3) << 3)) & 0xff;
  if(ref = 1)
    c += 1; /* XXX $B$&$^$/$$$C$F$$$J$$(B */
  buf[0] = 1;
  num & 0xffff;
  buf[1] = num & 0xff;
  buf[2] = (num >> 8) & 0xff;
  sendbytes(fd, buf, 3);

  if(readchar(fd) != 'A') {
    fprintf(stderr, "Reading AD failed(2).\n");
    return;
  }
  if(readchar(fd) != 0) {
    fprintf(stderr, "Reading AD failed(3).\n");
    return;
  }
    
  for(i=0; i<num; i++) {
    c = readchar(fd);
    if(c>=0)
      data[i] = c;
    else {
      fprintf(stderr, "Reading AD failed(4).\n");
      break;
    }
  }
  return(i);
}

int readEEPROM(int fd, int start, int count, unsigned char data[])
{
  char s, buf[4];
  int c, i;

  s = 'r';
  write(fd, &s, 1);
  if(readchar(fd) != s)
    return(-1);
  start &= 0xffff;
  count &= 0xffff;
  buf[0] = start & 0xff;
  buf[1] = (start >> 8) & 0xff;
  buf[2] = count & 0xff;
  buf[3] = (count >> 8) & 0xff;
  sendbytes(fd, buf, 4);

  if(readchar(fd) != 'R')
    return(-2);

  for(i=0; i<count; i++) {
    c = readchar(fd);
    if(c<0)
      break;
    s = data[i] = c;
    write(fd, &s, 1);
  }
  c = readchar(fd); /* dummy for end */
  if(c>=0)
    write(fd, &s, 1);
  return(i);
}

int writeEEPROM(int fd, int start, int count, unsigned char data[])
{
  char s, buf[4];
  int c, i;

  s = 'w';
  write(fd, &s, 1);
  if(readchar(fd) != s)
    return(-1);
  start &= 0xffff;
  count &= 0xffff;
  buf[0] = start & 0xff;
  buf[1] = (start >> 8) & 0xff;
  buf[2] = count & 0xff;
  buf[3] = (count >> 8) & 0xff;
  sendbytes(fd, buf, 4);

  if(readchar(fd) != 'W')
    return(-2);

  for(i=0; i<count; i++) {
    write(fd, &data[i], 1);
    c = readchar(fd);
    if(c<0 || c != data[i])
      break;
  }
  s = 0;
  write(fd, &s, 1); /* dummy write */
  c = readchar(fd);
  return(i);
}

void printLOGc(unsigned char buf[])
{
  if(buf[0] != 'A' || buf[1] != 'D') {
    printf("Invalid configuration.\n");
    return;
  }
  if(buf[2] & 0x80)
    printf("Logging is on after reset.\n");
  printf("0x%x\n", buf[3]);
  printf("Log period %d[s].\n", (buf[4] + (buf[5]<<8) + (buf[6]<<16) + (buf[7]<<24))/20000);
  printf("EEPROM writing start address 0x%x.\n", buf[8] + (buf[9]<<8));
  printf("EEPROM writing end address 0x%x.\n", buf[10] + (buf[11]<<8));
  printf("EEPROM current writing address 0x%x.\n", buf[12] + (buf[13]<<8));
}

int setLOGc(int fd, int port, unsigned int period, int romsize)
{
  unsigned char buf[0x10];

  buf[0] = 'A';
  buf[1] = 'D';
  buf[2] = 0x80; /* Logging mode on */
  buf[3] = port;
  period *= 20000;
  buf[4] = period & 0xff;
  buf[5] = (period>>8) & 0xff;
  buf[6] = (period>>16) & 0xff;
  buf[7] = (period>>24) & 0xff;
  buf[8] = 0x10;  /* EEPROM data start address */
  buf[9] = 0x0;
  romsize--;
  buf[10] = romsize & 0xff;
  buf[11] = (romsize>>8) & 0xff;
  buf[12] = 0x10; /* EEPROM next data writing address */
  buf[13] = 0x0;
  buf[14] = buf[15] = 0xff; /* dummy */

  return(writeEEPROM(fd, 0, 16, buf));
}

void printAINn(int fd, int ch, int ref, int num)
{
  unsigned char *p;
  int i, j;

  if((p = (unsigned char *)malloc(num)) == (unsigned char *)NULL) {
    fprintf(stderr, "malloc() failed.\n");
    return;
  }
  j = readAINn(fd, ch, ref, num, p);

  for(i=0; i<j; i++)
    printf("%d\n", p[i]);
  free(p);
}

void hexdump(unsigned char buf[], int num)
{
  int i;

  for(i=0; i<num; i++) {
    if((i % 16)==0)
      printf("\n");
    printf("%x ", buf[i]);
  }
  printf("\n");
}

void setRealMode(int fd, int port, unsigned int period)
{
  int c, i;
  char s, buf[5];

  s = 'd';
  write(fd, &s, 1);
  if(readchar(fd) != s) {
    fprintf(stderr, "Set real mode failed(1).\n");
    return;
  }
  buf[0] = port;
  period *= 25000;
  buf[1] = period & 0xff;
  buf[2] = (period>>8) & 0xff;
  buf[3] = (period>>16) & 0xff;
  buf[4] = (period>>24) & 0xff;
  sendbytes(fd, buf, 5);

  if(readchar(fd) != 'D') {
    fprintf(stderr, "Set real mode failed(2).\n");
    return;
  }
}

#ifdef DEBUG
int main( int argc, char *argv[] )
{
  int i;

#ifdef NEED_NODELAY
  if((asy = open(DEVICE, O_RDWR | O_NDELAY))<0) {
#else
  if((asy = open(DEVICE, O_RDWR))<0) {
#endif
    fprintf(stderr, "Colud not open serial port.\n");
    exit(2);
  }
  printf("Trying 38400bps\n");
  asy_init(asy, B38400);
  if(printmode() != 0) {
    printf("Trying 19200bps\n");
    asy_init(asy, B19200);
    if(printmode() == 0) {
      printDIN(asy);
      if(setbaudrate(asy, 9600) !=0)
	fprintf(stderr, "Set baud rate failed.\n");
    }
    printf("Trying 9600bps\n");
    asy_init(asy, B9600);
    printmode();
    printDIN(asy);
    for(i=0; i<4; i++) {
      printAIN(asy, i, 0);
    }
    for(i=0; i<3; i++)
      printAIN(asy, i, 1);

    printAINn(asy, 0, 0, 10);
    
    {
      unsigned char buf[30];
      if((i = readEEPROM(asy, 0, 30, buf))<0)
	fprintf(stderr, "EEPROM read failed(%d).\n", i);
      else {
	hexdump(buf, 20);
	printLOGc(buf);
      }
      i = setLOGc(asy, 0x8c, 300, 0x2000);
      if(i<0)
	fprintf(stderr, "EEPROM write failed(%d).\n", i);
      else
	fprintf(stderr, "EEPROM %d bytes worte.\n", i);
    }
/*     setRealMode(asy, 0x8c, 1); */
    if((i = setbaudrate(asy, 19200)) != 0)
      	fprintf(stderr, "Set baud rate failed(%d).\n", i);
  }

#if 0
  ioinit();
  atexit(ioreinit);
#endif
}
#else

int main( int argc, char *argv[] )
{
  int i, j;
  int autobaud = 0;
  int update = 0;
  float amp[] = {1.0, 1.0, 1.0, 1.0};
  float bias[] = {0.0, 0.0, 0.0, 0.0};

#ifdef NEED_NODELAY
  if((asy = open(DEVICE, O_RDWR | O_NDELAY))<0) {
#else
  if((asy = open(DEVICE, O_RDWR))<0) {
#endif
    fprintf(stderr, "Colud not open serial port.\n");
    exit(2);
  }
  
  for(i=1; i<argc; i++) {
    if(strcmp(argv[i], "-auto") == 0) {
      autobaud = 1;
      continue;
    }
    if(strcmp(argv[i], "-u") == 0 && (i+1)<argc) {
      update = atoi(argv[i+1]);
      i++;
      continue;
    }
    if(strncmp(argv[i], "-a", 2) == 0 && (i+1)<argc) {
      j = atoi(argv[i]+2);
      amp[j] = atof(argv[i+1]);
      i++;
      continue;
    }
    if(strncmp(argv[i], "-b", 2) == 0 && (i+1)<argc) {
      j = atoi(argv[i]+2);
      bias[j] = atof(argv[i+1]);
      i++;
      continue;
    }
  }

  if(autobaud) {
    printf("Trying 9600bps\n");
    asy_init(asy, B9600);
    if(setPCMODE(asy)!=0) {
      printf("Trying 19200bps\n");
      asy_init(asy, B19200);
      if(setPCMODE(asy)!=0) {
	printf("Trying 38400bps\n");
	asy_init(asy, B38400);
	if(setPCMODE(asy)!=0) {
	  printf("Aborted.\n");
	  exit(1);
	}
      }
    }
  } else {
    asy_init(asy, B9600);
    if(setPCMODE(asy) !=0)
      fprintf(stderr, "Set ADC mode failed.\n");
  }

  if(update==0) {
    float f;
    for(j=0; j<4; j++) {
      if(amp[j] == 0.0)
	continue;
      f = readAIN(asy, j, 0) * amp[j] + bias[j];
      printf("%3.1f\t", f);
    }
    putchar('\n');
  } else
    while(1) {
      float f;
      for(j=0; j<4; j++) {
	if(amp[j] == 0.0)
	  continue;
	f = readAIN(asy, j, 0) * amp[j] + bias[j];
	printf("%3.1f\t", f);
      }
      putchar('\n');
      sleep(update);
    }
  
  close(asy);
}
#endif
